# frozen_string_literal: true

require "reline"
require "irb"

require_relative "rubic/version"

# The Rubic REPL. You probably want to start a session with `Rubic::run`
module Rubic
  def self.run
    Rubic.new.eval_forever
  end

  # Represents program (as a set of lines)
  class LineGroup
    def initialize
      @lines = {}
    end

    def lines
      @lines.sort_by { |line_num, _| line_num }
    end

    def clear
      @lines = {}
    end

    def update(line_num, expr)
      if expr.nil?
        @lines.delete(line_num)
      else
        @lines[line_num] = expr
      end
    end

    def resize(gap)
      @lines =
        @lines
        .sort_by { |line_num, _| line_num }
        .each_with_index
        .map { |hsh, i| [(i + 1) * gap, hsh[1]] }
        .to_h
    end

    def stringify(with_line_numbers)
      lines = @lines.map { |k, v| [k, v] }

      lines.sort_by { |k, _| k }.map do |line_num, line|
        if with_line_numbers
          "#{line_num} #{line}"
        else
          line
        end
      end.join("\n")
    end

    def save(file_name, with_line_numbers)
      File.open(file_name, "w") do |file|
        file.write(stringify(with_line_numbers))
        file.write("\n")
      end
    end

    def load(file_name, with_line_numbers)
      @lines = File.readlines(file_name, chomp: true)
                   .each_with_index
                   .map do |line, i|
        if with_line_numbers
          line_num, line = line.split(" ", 2)
          line_num = Integer(line_num)

          [line_num, line]
        else
          [i + 1, line]
        end
      end.to_h
    end
  end

  # The REPL itself
  class Rubic
    def initialize
      @workspace = IRB::WorkSpace.new(TOPLEVEL_BINDING.dup)
      @lines = LineGroup.new
    end

    def eval_forever
      # so we can reset settings after
      # todo need to put more in here (maybe? IRB doesn't...)
      history_len = Reline::HISTORY.size
      og_completion_proc = Reline.completion_proc
      og_autocompletion = Reline.autocompletion

      Reline.completion_proc = lambda { |target, preposting, postposting|
        completion(target, preposting, postposting)
      }

      Reline.autocompletion = true
      Reline.completer_quote_characters = ""
      Reline.output_modifier_proc = proc do |input, complete:|
        output_modifier(input, complete)
      end

      loop do
        line = Reline.readline("> ")
        break if line.nil?

        Reline::HISTORY.push(line) if line.strip != ""

        cmd, args = line.split(" ", 2)
        case cmd
        when "list"
          @lines.lines.each do |line_num, line|
            puts output_modifier("#{line_num} #{line}", false)
          end
        when "run"
          run(@lines.stringify(false))
        when "clear"
          @lines.clear
        when "load"
          if args.nil? || args.empty?
            puts "Usage: load filename.rub"
          else
            @lines.load(args, true)
          end
        when "save"
          if args.nil? || args.empty?
            puts "Usage: save filename.rub"
          else
            @lines.save(args, true)
          end
        when "import"
          if args.nil? || args.empty?
            puts "Usage: import filename.rb"
          else
            @lines.load(args, false)
          end
        when "export"
          if args.nil? || args.empty?
            puts "Usage: export filename.rb"
          else
            @lines.save(args, false)
          end
        when "squish"
          @lines.resize(1)
        when "widen"
          @lines.resize(10)
        when "help"
          print_help
        when "exit", "bye"
          break
        else
          line_num, expr = line.split(" ", 2)
          line_num = begin
            Integer(line_num)
          rescue StandardError
            nil
          end

          if line_num.nil?
            run(line) if line != ""
          else
            @lines.update(line_num, expr)
          end
        end
      rescue Interrupt
        puts "^C"
      end

      # so that the shell is put on a new line
      # if we exit with ctrl+d
      puts
    ensure
      # reset history
      delta = Reline::HISTORY.size - history_len
      (0...delta).each do
        Reline::HISTORY.pop
      end

      Reline.completion_proc = og_completion_proc
      Reline.autocompletion = og_autocompletion
    end

    private

    def completion(target, preposting, postposting)
      full_line = "#{preposting}#{target}"

      line_num, expr = full_line.split(" ", 2)
      line_num = begin
        Integer(line_num)
      rescue StandardError
        nil
      end

      target = expr || "" unless line_num.nil?

      if line_num.nil? && %w[load save import export].any? { |cmd| full_line.start_with?("#{cmd} ") }
        _, path = full_line.split(" ", 2)

        Dir["#{path}*"]
      else
        completor = IRB::RegexpCompletor.new
        candidates = completor.completion_candidates(preposting, target, postposting, bind: @workspace.binding)

        if line_num.nil? && preposting == ""
          candidates += %w[list run load save import export squish widen help exit bye]
        end

        candidates
      end
    end

    def output_modifier(input, complete)
      line_num, expr = input.split(" ", 2)
      line_num = begin
        Integer(line_num)
      rescue StandardError
        nil
      end

      lvars = @workspace.send(:local_variables) # TODO: hacky!!

      if line_num.nil?
        IRB::Color.colorize_code(input, complete: complete, local_variables: lvars)
      else
        context = @lines.lines.take_while { |ln, _| ln < line_num }.map { |_, line| line }
        context_len = context.size
        context << expr
        context = context.join("\n")

        num = IRB::Color.colorize_code(line_num, complete: complete, local_variables: lvars)
        expr = IRB::Color.colorize_code(context, complete: complete, local_variables: lvars).split("\n")
        expr.shift(context_len)
        expr = expr.join("\n")

        "#{num} #{expr}"
      end
    end

    def run(expr)
      is_exception, ret = begin
        [false, @workspace.evaluate(expr)]
      rescue StandardError, SyntaxError => e
        [true, e]
      end

      if is_exception
        pp ret
        puts ret.backtrace
      else
        print "=> "
        pp ret
      end
    end

    def print_help
      puts "
Usage:
\thelp\t\tDisplay this help message
\t{num} {expr}\tSet line {num} to {expr}
\t{num}\t\tClear line {num}
\tlist\t\tList current program
\trun\t\tRun current program
\tclear\t\tClear current program
\tsquish\t\tRe-numbers lines so that they go 1, 2, 3, etc.
\twiden\t\tRe-numbers lines so that they go 10, 20, 30, etc.
\tload {file}\tLoad program from file (expects line numbers in the file)
\tsave {file}\tSave program to file (includes line numbers in file)
\timport {file}\tImports a \"normal\" Ruby program (no line numbers)
\texport {file}\tExports a \"normal\" Ruby program (no line numbers)
"
    end
  end
end
